#!/usr/bin/env python
#
#   ConVirt   -  Copyright (c) 2008 Convirture Corp.
#   ======
#
# ConVirt is a Virtualization management tool with a graphical user
# interface that allows for performing the standard set of VM operations
# (start, stop, pause, kill, shutdown, reboot, snapshot, etc...). It
# also attempts to simplify various aspects of VM lifecycle management.
#
#
# This software is subject to the GNU General Public License, Version 2 (GPLv2)
# and for details, please consult it at:
#
#    http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt
# 
#
# author : Jd <jd_jedi@users.sourceforge.net>
#

from convirt.core.utils.utils import PyConfig
import re, os

from convirt.core.utils.NodeProxy import Node
from convirt.core.utils.constants import *



# Contains base classes representing
#        VM,
#        VMConfig,
#        VMStats,
#        VMOperationsException
# Also, has utility classes
#        DiskEntry
#        NetworkEntry

class VMOperationException(Exception):
   def __init__(self, message, errno = None):
      if errno:
         Exception.__init__(self, (errno, message))
      else:
         Exception.__init__(self,  message)
      self.message = message
      self.errno = errno

   def __repr__(self):
      if self.errno != None:
         return "[Error %s]: %s" % (str(self.errno), self.message)
      else:
         return self.message
      


class VM:
    """
    This represent Doms. It encapsulates information about
    running virtual machine : state as well as resource stats
    """
    RUNNING  = 0
    BLOCKED  = 1
    PAUSED   = 2
    SHUTDOWN = 3
    CRASHED  = 4
    NOT_STARTED    = -1
    UNKNOWN  = -2


    def __init__(self, node, config=None,vm_info=None):
       if not config and not vm_info:
          raise Exception("config and vm_info both can not be None")
       
       self.node = node
       self._config = config
       self._vm_info= vm_info

       self.id = None
       self.name = None
       self._is_resident = False
       self.state  = self.UNKNOWN
       
       # Now call the init method that would initialized the above
       # state variables
       self.init()

       
    # to be provided by the derived class
    # set id, name, and _is_resident state
    def init(self):
       raise Exception("Must be implemented by derived class.")


    def get_platform(self):
       raise Exception("Platform not defined!!" + str(self.__class__))

    # interface methods
    """ return the state in which the VM is """
    def get_state(self):
       return self.state

    """ Is the vm dom0 """
    def isDom0(self):
       return False # for most platforms

    def isDomU(self):
       return True # for most platforms

    """ refresh the vm information """
    def refresh(self):
       try:
          self._vm_info = self.node.get_vmm().refresh(self.id)
          self.init()
       except VMOperationException, ex:
          self._vm_info = None
          self._is_resident = False

    """ Save the vm to a given file """
    def _save(self, filename):
       self.node.get_vmm().save(self.id, filename)


    def _reboot(self):
       self.node.get_vmm().reboot(self.id)
       self._vm_info =  None
       self._is_resident = False


    def _start(self):
       self.node.get_vmm().start(self._config)
       self.state = self.RUNNING

    def _shutdown(self):
       self.node.get_vmm().shutdown(self.id)
       self._vm_info =  None
       self._is_resident = False
       self.state = self.SHUTDOWN


    def _destroy(self):
       self.node.get_vmm().destroy(self.id)
       self._vm_info =  None
       self._is_resident = False
       self.state = self.NOT_STARTED


    def _pause(self):
       self.node.get_vmm().pause(self.id)
       self.state = self.PAUSED

    def _unpause(self):
       self.node.get_vmm().unpause(self.id)
       self.state = self.RUNNING

    def _suspend(self):
       self.node.get_vmm().suspend(self.id)
       self.state = PAUSED

    def _resume(self):
       self.node.get_vmm().resume(self.id)
       self.state = RUNNING

    def _migrate(self, dest,live, port):
       self.node.get_vmm().migrate(self.id, dest, live, port)
       self.state = self.NOT_STARTED


    def is_resident(self):
       return self._is_resident
    
    # change to live running VM
    def setVCPUs(self, value):
       self.node.get_vmm().setVCPUs(self.id, value)


    def setMem(self,value):
       self.node.get_vmm().setMem(self.id, value)

    # Associate a config
    def get_config(self):
       return self._config 
    
    def set_config(self, config):
       self._config=config
       pass

    def get_info(self):
       return self._vm_info

    # get some metric information
    # TBD: Needs some standardization on what minimal info 
    def get_snapshot(self):
       pass


    # expose the vm details via getitem
    # Need some standardization so that UI can display it.
    def __getitem__(self, param):
       pass

    # display strings
    def get_state_string(self):
       state = self.get_state()
       if state == self.RUNNING: return "Running"
       elif state ==  self.BLOCKED : return "Blocked"
       elif state ==  self.PAUSED : return "Paused"
       elif state == self.SHUTDOWN: return "Shutdown"
       elif state == self.CRASHED: return "Crashed"
       elif state == self.UNKNOWN: return "Unknown"
        
       return "Unknown"

    # return (cmd, args) for the console
    def get_console_cmd(self):
       pass

    # get the vnc port to be used
    def get_vnc_port(self):
       return None

    # should we use the graphical display or console
    def is_graphical_console(self):
       return False


    def is_hvm(self):
       if self._config is None:
          return None
       else:
          return self._config.is_hvm()
       
# Disk entry class : very generic.
class DiskEntry:
    """This class provides a structured representation of the disk
    string. The following attributes are available:

     type:     {file|phy}
     filename: the path to the VBD or device
     device:   the device to be used in the guest system
     mode:     {r|w|w!}
     """

    def __init__(self, input):
        if type(input) is str:
            m = re.match("^(.*):(.*),(.*),(.*)$", input)
            if m:
                self.type = m.group(1)
                self.filename = m.group(2)
                self.device = m.group(3)
                self.mode = m.group(4)
            else:
                print "disk entry : No regexp match for", input
                raise Exception("could not parse disk "+ input)
        elif type(input) in (list, tuple):
            self.type, self.filename, self.device, self.mode = input

    def __repr__(self):
        return "%s:%s,%s,%s" % (self.type, self.filename, self.device,
                                self.mode)

# Network class
#class NetworkEntry:



# DomConfig object, will use python config 
class VMConfig(PyConfig):
    """
    reprsents startup config object (information in the conf file)
    """
    
    # Module variable.
    imps = ["import sys",
            "import os",
            "import os.path",
            ]
    

    signature = "# Automtically generated by ConVirt\n"



    # DomConfig follows
    def __init__(self, node=None, filename = None):
        """
        read stuff from file and populate the config
        when filename is None, creates an empty config
        """
        self.node = node
        self.convirt_generated = False

        
        PyConfig.__init__(self,
                          node,
                          filename,
                          VMConfig.signature)

        if filename is None : return

        if len(self.lines) > 0:
            if self.lines[0].find(self.signature) >= 0:
                    self.convirt_generated = True
            
        if self["name"] is None:
            raise Exception("No dom name specified")

        #pick up the name from the file
        self.name = self["name"]

    
    # read_conf
    def read_conf(self, init_glob=None, init_locs=None):
        # Leverage the fact that conf files are python syntax.
        # save on parsing code
        
        # Create global and local dicts for the file.
        # Initialize locals to the vars.
        globs = {}
        locs = {}

        cmd = '\n'.join(self.imps)  

        # Use exec to do the standard imports and
        # define variables we are passing to the script.
        exec cmd in globs, locs
        return PyConfig.read_conf(self, globs, locs)
    
    def __setitem__(self, name, item):
        self.options[name] = item
        if name == "name":
            self.name = item

    # try to keep them in sync
    def set_name(self, name):
        self.name = name
        self["name"] = name
    
    
    # get the configured disks
    def getDisks(self):
        """
        This method returns a more structured version of the config file's
        disk entry
        """
        reslist = []
        for str in self["disk"]:
            reslist.append(DiskEntry(str))
        return reslist


    # Assumes that the file read contains single line values
    # now preserves the comments and order
    def write(self):
        """Writes the settings out to the filename specified during
        initialization"""
        self.name = self["name"]
        PyConfig.write(self)
        self.convirt_generated = True


    def is_convirt_generated(self):
        return self.convirt_generated

    # Get raw file content    
    def get_contents(self):
        f = self.node.node_proxy.open(filename)
        lines = f.readlines()
        f.close()
        contents = "".join(lines)
        return contents

    # Write raw contents
    def write_contents(self, contents):
        """Writes the settings out to the filename specified during
        initialization"""
        outfile = self.node.node_proxy.open(self.filename, 'w')
        outfile.write(contents)
        outfile.close()



    # Validate the config. Actual details to be supplied by the
    # derived class
    
    def validate(self):
        """Attempts to validate that the settings are not going to lead to
        any errors when the dom is started, and returns a list of the
        errors as strings"""

        result = []
        return result

    def is_hvm(self):
       if self["builder"] == "hvm" :
          return True
        

class VMStats:
    """
    represents statatistics/measurements for a dom. (CPU, I/O etc)
    This is abstracted out so we can cut over to some other source for
    runtime statastics/measurements
    """

    def __init__(self, vm):
        """
        constructor, dom for which the stats are to be obtained.
        """
        self.vm = vm
        self.stat = {}

    def get_snapshot(self):
        # to be implemented by derived class
        return self.stat
